@@ -1,4 +1,4 @@
-# Copyright 2021 The Brax Authors.
+# Copyright 2022 The Brax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,6 +14,7 @@
 
 """Input normalization utils."""
 
+import functools
 from typing import Any
 
 import jax
@@ -36,12 +37,34 @@ def synchronize_hosts():
 
 
 def _fingerprint(x: Any) -> float:
-  sums = jax.tree_map(jnp.sum, x)
+  sums = jax.tree_util.tree_map(jnp.sum, x)
   return jax.tree_util.tree_reduce(lambda x, y: x + y, sums)
 
 
-def is_synchronized(x: Any, axis_name: str) -> jnp.ndarray:
+def is_replicated(x: Any, axis_name: str) -> jnp.ndarray:
+  """Returns whether x is replicated.
+
+  Should be called inside a function pmapped along 'axis_name'
+  Args:
+    x: Object to check replication.
+    axis_name: pmap axis_name.
+
+  Returns:
+    boolean whether x is replicated.
+  """
   fp = _fingerprint(x)
   return jax.lax.pmin(
       fp, axis_name=axis_name) == jax.lax.pmax(
           fp, axis_name=axis_name)
+
+
+def assert_is_replicated(x: Any, debug: Any = None):
+  """Returns whether x is replicated.
+
+  Should be called from a non-jitted code.
+  Args:
+    x: Object to check replication.
+    debug: Debug message in case of failure.
+  """
+  f = functools.partial(is_replicated, axis_name='i')
+  assert jax.pmap(f, axis_name='i')(x)[0], debug
